/*
   shadow-utils

   su-selinux.c

   Copyright (C) Jakub Hrozek <jhrozek@redhat.com>        2010
   Copyright (C) Peter Vrabec <pvrabec@redhat.com>        2011

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <config.h>

#include "defines.h"

#include <stdio.h>
#include <selinux/selinux.h>
#include <semanage/semanage.h>


#ifndef DEFAULT_SERANGE
#define DEFAULT_SERANGE "s0"
#endif


static void semanage_error_callback(void *varg,
                                    semanage_handle_t *handle,
                                    const char *fmt, ...)
{
    int ret;
    char * message = NULL;
    va_list ap;


    va_start(ap, fmt);
    ret = vasprintf(&message, fmt, ap);
    va_end(ap);
    if (ret < 0) {
        /* ENOMEM */
        return;
    }

    switch (semanage_msg_get_level(handle)) {
        case SEMANAGE_MSG_ERR:
        case SEMANAGE_MSG_WARN:
	    fprintf(stderr, "[libsemanage]: %s\n", message);
            break;
        case SEMANAGE_MSG_INFO:
	    /* nop */
            break;
    }

    free(message);
}


static semanage_handle_t *semanage_init(void)
{
    int ret;
    semanage_handle_t *handle = NULL;

    handle = semanage_handle_create();
    if (!handle) {
	fprintf(stderr, _("Cannot create SELinux management handle\n"));
        return NULL;
    }

    semanage_msg_set_callback(handle, semanage_error_callback, NULL);

    ret = semanage_is_managed(handle);
    if (ret != 1) {
	fprintf(stderr, _("SELinux policy not managed\n"));
        goto fail;
    }

    ret = semanage_access_check(handle);
    if (ret < SEMANAGE_CAN_READ) {
	fprintf(stderr, _("Cannot read SELinux policy store\n"));
        goto fail;
    }

    ret = semanage_connect(handle);
    if (ret != 0) {
	fprintf(stderr, _("Cannot estabilish SELinux management connection\n"));
        goto fail;
    }

    ret = semanage_begin_transaction(handle);
    if (ret != 0) {
	fprintf(stderr, _("Cannot begin SELinux transaction\n"));
        goto fail;
    }

    return handle;
fail:
    semanage_handle_destroy(handle);
    return NULL;
}


static int semanage_user_mod(semanage_handle_t *handle,
                             semanage_seuser_key_t *key,
                             const char *login_name,
                             const char *seuser_name)
{
    int ret;
    semanage_seuser_t *seuser = NULL;

    semanage_seuser_query(handle, key, &seuser);
    if (seuser == NULL) {
	fprintf(stderr, _("Could not query seuser for %s\n"), login_name);
        ret = 1;
        goto done;
    }

    ret = semanage_seuser_set_mlsrange(handle, seuser, DEFAULT_SERANGE);
    if (ret != 0) {
	fprintf(stderr, _("Could not set serange for %s\n"), login_name);
        ret = 1;
        goto done;
    }

    ret = semanage_seuser_set_sename(handle, seuser, seuser_name);
    if (ret != 0) {
	fprintf(stderr, _("Could not set sename for %s\n"), login_name);
        ret = 1;
        goto done;
    }

    ret = semanage_seuser_modify_local(handle, key, seuser);
    if (ret != 0) {
	fprintf(stderr, _("Could not modify login mapping for %s\n"), login_name);
        ret = 1;
        goto done;
    }

    ret = 0;
done:
    semanage_seuser_free(seuser);
    return ret;
}


static int semanage_user_add(semanage_handle_t *handle,
                             semanage_seuser_key_t *key,
                             const char *login_name,
                             const char *seuser_name)
{
    int ret;
    semanage_seuser_t *seuser = NULL;

    ret = semanage_seuser_create(handle, &seuser);
    if (ret != 0) {
	fprintf(stderr, _("Cannot create SELinux login mapping for %s\n"), login_name);
        ret = 1;
        goto done;
    }

    ret = semanage_seuser_set_name(handle, seuser, login_name);
    if (ret != 0) {
	fprintf(stderr, _("Could not set name for %s\n"), login_name);
        ret = 1;
        goto done;
    }

    ret = semanage_seuser_set_mlsrange(handle, seuser, DEFAULT_SERANGE);
    if (ret != 0) {
	fprintf(stderr, _("Could not set serange for %s\n"), login_name);
        ret = 1;
        goto done;
    }

    ret = semanage_seuser_set_sename(handle, seuser, seuser_name);
    if (ret != 0) {
	fprintf(stderr, _("Could not set SELinux user for %s\n"), login_name);
        ret = 1;
        goto done;
    }

    ret = semanage_seuser_modify_local(handle, key, seuser);
    if (ret != 0) {
	fprintf(stderr, _("Could not add login mapping for %s\n"), login_name);
        ret = 1;
        goto done;
    }

    ret = 0;
done:
    semanage_seuser_free(seuser);
    return ret;
}


int set_seuser(const char *login_name, const char *seuser_name)
{
    semanage_handle_t *handle = NULL;
    semanage_seuser_key_t *key = NULL;
    int ret;
    int seuser_exists = 0;

    if (seuser_name == NULL) {
        /* don't care, just let system pick the defaults */
        return 0;
    }

    handle = semanage_init();
    if (!handle) {
	fprintf(stderr, _("Cannot init SELinux management\n"));
        ret = 1;
        goto done;
    }

    ret = semanage_seuser_key_create(handle, login_name, &key);
    if (ret != 0) {
	fprintf(stderr, _("Cannot create SELinux user key\n"));
        ret = 1;
        goto done;
    }

    ret = semanage_seuser_exists(handle, key, &seuser_exists);
    if (ret < 0) {
	fprintf(stderr, _("Cannot verify the SELinux user\n"));
        ret = 1;
        goto done;
    }

    if (seuser_exists) {
        ret = semanage_user_mod(handle, key, login_name, seuser_name);
        if (ret != 0) {
	    fprintf(stderr, _("Cannot modify SELinux user mapping\n"));
            ret = 1;
            goto done;
        }
    } else {
        ret = semanage_user_add(handle, key, login_name, seuser_name);
        if (ret != 0) {
	    fprintf(stderr, _("Cannot add SELinux user mapping\n"));
            ret = 1;
            goto done;
        }
    }

    ret = semanage_commit(handle);
    if (ret < 0) {
	fprintf(stderr,_("Cannot commit SELinux transaction\n"));
        ret = 1;
        goto done;
    }

    ret = 0;

done:
    semanage_seuser_key_free(key);
    semanage_handle_destroy(handle);
    return ret;
}





int del_seuser(const char *login_name)
{
    semanage_handle_t *handle = NULL;
    semanage_seuser_key_t *key = NULL;
    int ret;
    int exists = 0;

    handle = semanage_init();
    if (!handle) {
	fprintf(stderr, _("Cannot init SELinux management\n"));
        ret = 1;
        goto done;
    }

    ret = semanage_seuser_key_create(handle, login_name, &key);
    if (ret != 0) {
	fprintf(stderr, _("Cannot create SELinux user key\n"));
        ret = 1;
        goto done;
    }

    ret = semanage_seuser_exists(handle, key, &exists);
    if (ret < 0) {
	fprintf(stderr, _("Cannot verify the SELinux user\n"));
        ret = 1;
        goto done;
    }

    if (!exists) {
        fprintf(stderr, _("Login mapping for %s is not defined, OK if default mapping was used\n"), 
		       login_name);
        ret = 0;  /* probably default mapping */
        goto done;
    }

    ret = semanage_seuser_exists_local(handle, key, &exists);
    if (ret < 0) {
	fprintf(stderr, _("Cannot verify the SELinux user\n"));
        ret = 1;
        goto done;
    }

    if (!exists) {
	fprintf(stderr, _("Login mapping for %s is defined in policy, cannot be deleted\n"), 
		       login_name);
        ret = 0; /* Login mapping defined in policy can't be deleted */
        goto done;
    }

    ret = semanage_seuser_del_local(handle, key);
    if (ret != 0) {
	fprintf(stderr, _("Could not delete login mapping for %s"), login_name);
        ret = 1;
        goto done;
    }

    ret = semanage_commit(handle);
    if (ret < 0) {
	fprintf(stderr, _("Cannot commit SELinux transaction\n"));
        ret = 1;
        goto done;
    }

    ret = 0;
done:
    semanage_handle_destroy(handle);
    return ret;
}

